# -*- coding: utf-8 -*-

import os
import json
import time
import string
import random

try:
    from urllib.parse import parse_qs, urlparse, urlencode
except ImportError:
    from urlparse import parse_qs, urlparse
    from urllib import urlencode

import requests
from requests.exceptions import (
    ConnectTimeout,
    ConnectionError,
    ReadTimeout,
    HTTPError,
    RequestException
)

from bingads.authorization import OAuthDesktopMobileAuthCodeGrant

from bing_content_api.exception import *
from bing_content_api import logger

VERSION = 'v9.1'
BMC_HOST = "https://content.api.bingads.microsoft.com"
BMC_HOST_V2 = "https://content.api.ads.microsoft.com"
BMC_HOST_V2_RESOURCES = ['productstatuses']

PRODUCTION = 'production'
SANDBOX = 'sandbox'

OAuthService = dict()
OAuthService[PRODUCTION] = {
    'REDIRECTION_URI': 'https://login.live.com/oauth20_desktop.srf',
    'AUTH_TOKEN_URI': 'https://login.live.com/oauth20_token.srf',
    'AUTHORIZE_URI': 'https://login.live.com/oauth20_authorize.srf?client_id={0}&scope=bingads.manage&response_type={1}&redirect_uri={2}',
    'SCOPE': 'msads.manage'
}
OAuthService[SANDBOX] = {
    'REDIRECTION_URI': 'https://login.live-int.com/oauth20_desktop.srf',
    'AUTH_TOKEN_URI': 'https://login.live-int.com/oauth20_token.srf',
    'AUTHORIZE_URI': 'https://login.live-int.com/oauth20_authorize.srf?client_id={0}&scope=bingads.manage&response_type={1}&redirect_uri={2}&prompt=login',
    'SCOPE': 'msads.manage'
}


class Connection(object):
    def __init__(
            self, client_id, developer_token, merchant_id, refresh_token, environment=PRODUCTION,
            catalog_id=None, max_retries=12, timeout=60):
        self.max_retries = max_retries
        self.timeout = timeout
        self._last_response = None
        self.client_id = client_id
        self.developer_token = developer_token
        self.merchant_id = merchant_id
        self.catalog_id = catalog_id
        self.refresh_token = refresh_token
        self.auth = None

        if environment not in [PRODUCTION, SANDBOX]:
            raise ValueError('[UnknownEnvironment] {}'.format(environment))
        self.environment = environment

        try:
            self.state = ''.join(random.choices(string.ascii_uppercase + string.digits, k=10))
        except:
            self.state = ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(10))

        self.token = None
        if refresh_token:
            try:
                self.authenticate()
            except:
                raise

    def authenticate(self):
        if not self.refresh_token:
            raise ValueError('You must supply a valid refresh token to use bing shopping')

        self.auth = OAuthDesktopMobileAuthCodeGrant(
            client_id=self.client_id,
            env=self.environment
        )

        self.auth.state = self.state

        self.token = self.auth.request_oauth_tokens_by_refresh_token(self.refresh_token)

    def full_path(self, url):
        if self.merchant_id:
            v2_res = any([url.find(res) != -1 for res in BMC_HOST_V2_RESOURCES])
            if v2_res:
                full_url = BMC_HOST_V2 + '/{}/bmc/stores/{}/{}'.format(VERSION, self.merchant_id, url)
            else:
                full_url = BMC_HOST + '/shopping/{}/bmc/{}/{}'.format(VERSION, self.merchant_id, url)
        else:
            full_url = BMC_HOST + '/shopping/{}/bmc/{}'.format(VERSION, url)

        # if self.catalog_id:
        #     full_url = full_url + '?bmc-catalog-id={}'.format(self.catalog_id)

        return full_url

    def make_request(self, method, url, data=None, params=None, headers=None):
        response = self._run_method(method, url, data, params, headers)
        return self._handle_response(url, response)

    def _run_method(self, method, url, data=None, query=None, headers=None):
        if not self.token or not self.token.access_token:
            raise RuntimeError('[AuthorizionError] Authorizision is required!')

        if headers is None:
            headers = {}
        headers.update({
            'DeveloperToken': self.developer_token,
            'AuthenticationToken': self.token.access_token,
            'Content-Type': 'application/json'
        })
        
        if url and url[:4] != "http":
            if url[0] == '/':  # can call with /resource if you want
                url = url[1:]
            url = self.full_path(url)
        elif not url:  # blank path
            url = self.full_path(url)

        if query is None:
            query = {}
        qs = urlencode(query)
        if qs:
            qs = "?" + qs
        url += qs

        logger.debug("%s %s" % (method, url, ))
        logger.debug('Headers: %s, data: %s', headers, data)
        # make and send the request
        retries = self.max_retries
        response = None
        while retries > 0:
            try:
                response = requests.request(
                    method, url, headers=headers, data=data, timeout=self.timeout)
                response.raise_for_status()
                break
            except (ConnectTimeout, ConnectionError, ReadTimeout):
                retries -= 1
                time.sleep(30)
                if retries <= 0:
                    raise
            except HTTPError as err:
                retries -= 1

                if err.response.status_code < 500:
                    break

                time.sleep(7 if 500 <= err.response.status_code < 600 else 1)
            except RequestException:
                break
            except Exception as e:
                raise e

        return response

    def _handle_response(self, url, res):
        self._last_response = res
        result = {}
        if res.status_code in (200, 201, 202):
            try:
                result = res.json()
            except Exception as e:  # json might be invalid, or store might be down
                e.message += " (_handle_response failed to decode JSON: " + str(res.content) + ")"
                raise  # TODO better exception
        elif res.status_code == 204:
            pass
        elif res.status_code >= 500:
            raise ServerException(
                "%d %s @ %s: %s" % (res.status_code, res.reason, url, res.content), res)
        elif res.status_code == 400:
            raise BadRequestException("%d %s @ %s: %s" % (res.status_code, res.reason, url, res.content), res)
        elif res.status_code == 401:
            raise UnauthorizedException(
                "%d %s @ %s: %s" % (res.status_code, res.reason, url, res.content), res)
        elif res.status_code == 404:
            raise ResourceNotFoundException("%d %s @ %s: %s" % (res.status_code, res.reason, url, res.content), res)
        elif res.status_code == 409:
            raise ResourceConflictException(
                "%d %s @ %s: %s" % (res.status_code, res.reason, url, res.content), res)
        elif res.status_code == 413:
            raise RequestEntityTooLargeException(
                "%d %s @ %s: %s" % (res.status_code, res.reason, url, res.content), res)
        elif res.status_code >= 400:
            raise ClientRequestException("%d %s @ %s: %s" % (res.status_code, res.reason, url, res.content), res)
        elif res.status_code >= 300:
            raise RedirectionException("%d %s @ %s: %s" % (res.status_code, res.reason, url, res.content), res)

        return result
