# -*- coding: utf-8 -*-

import json


class Mapping(dict):
    """
    Mapping

    provides '.' access to dictionary keys
    """
    def __init__(self, mapping, *args, **kwargs):
        """
        Create a new mapping. Filters the mapping argument
        to remove any elements that are already methods on the
        object.

        For example, Orders retains its `coupons` method, instead
        of being replaced by the dict describing the coupons endpoint
        """

        mapping = mapping or {}

        filter_args = {k: mapping[k] for k in mapping if k not in dir(self)}
        self.__dict__ = self
        dict.__init__(self, filter_args, *args, **kwargs)

    def __str__(self):
        """
        Display as a normal dict, but filter out underscored items first
        """
        return str({k: self.__dict__[k] for k in self.__dict__ if not k.startswith("_")})

    def __repr__(self):
        return "<%s at %s, %s>" % (type(self).__name__, hex(id(self)), str(self))


class ApiResource(Mapping):
    resource_name = ""  # The identifier which describes this resource in urls

    @classmethod
    def _create_object(cls, response, connection=None):
        if response and not isinstance(response, dict):
            return [cls(obj, _connection=connection) for obj in response]
        else:
            return cls(response, _connection=connection)

    @classmethod
    def _make_request(cls, method, url, connection, data=None, params=None, headers=None):
        return connection.make_request(method, url, data, params, headers)

    @classmethod
    def _get_path(cls, id):
        return "%s/%s" % (cls.resource_name, id)

    @classmethod
    def get(cls, id, connection=None, **params):
        response = cls._make_request('GET', cls._get_path(id), connection, params=params)
        return cls._create_object(response, connection=connection)


class ApiSubResource(ApiResource):
    parent_resource = ""
    parent_key = ""

    @classmethod
    def _get_path(cls, id, parentid):
        return "%s/%s/%s/%s" % (cls.parent_resource, parentid, cls.resource_name, id)

    @classmethod
    def get(cls, parentid, id, connection=None, **params):
        response = cls._make_request('GET', cls._get_path(id, parentid), connection, params=params)
        return cls._create_object(response, connection=connection)

    def parent_id(self):
        return self[self.parent_key]


class BatchableApiResource(ApiResource):
    @classmethod
    def _batch_path(cls):
        return '{}/batch'.format(cls.resource_name)

    @classmethod
    def batch(cls, data, connection=None):
        return cls._make_request(
            'POST', cls._batch_path(), connection=connection, data=json.dumps(data))


class CreateableApiResource(ApiResource):
    @classmethod
    def _create_path(cls):
        return cls.resource_name

    @classmethod
    def create(cls, data, connection=None):
        response = cls._make_request(
            'POST', cls._create_path(), connection, data=json.dumps(data))
        return cls._create_object(response, connection=connection)


class CreateableApiSubResource(ApiSubResource):
    @classmethod
    def _create_path(cls, parentid):
        return "%s/%s/%s" % (cls.parent_resource, parentid, cls.resource_name)

    @classmethod
    def create(cls, parentid, data, connection=None):
        response = cls._make_request(
            'POST', cls._create_path(parentid), connection, data=json.dumps(data))
        return cls._create_object(response, connection=connection)


class ListableApiResource(ApiResource):
    @classmethod
    def _get_all_path(cls):
        return cls.resource_name

    @classmethod
    def all(cls, connection=None, **params):
        """
            Returns first page if no params passed in as a list.
        """

        rsp = cls._make_request('GET', cls._get_all_path(), connection, params=params)
        if rsp and 'resources' in rsp and rsp['resources']:
            return cls._create_object(rsp['resources'], connection=connection)
        else:
            return [cls._create_object({}, connection=connection)]

    @classmethod
    def iterall(cls, connection=None, **kwargs):
        """
            Returns a autopaging generator that yields each object returned one by one.
        """
        start_token = None
        params = kwargs.copy()
        params['max-results'] = 250

        while True:
            if start_token:
                params['start-token'] = start_token
            else:
                params.pop('start-token', None)

            rsp = cls._make_request('GET', cls._get_all_path(), connection, params=params)
            if rsp:
                resources = rsp.get('resources', [])
                if not resources:
                    break

                for resource in resources:
                    yield cls._create_object(resource, connection=connection)

                # if 'kind' in rsp and rsp['kind'] == 'content#productsListResponse' and \
                #     'nextPageToken' in rsp and rsp['nextPageToken']:
                if 'nextPageToken' in rsp and rsp['nextPageToken']:
                    start_token = rsp['nextPageToken']
                    continue

                break
            else:
                yield cls._create_object({}, connection=connection)  # needed for case where there is no objects

            break


class ListableApiSubResource(ApiSubResource):
    @classmethod
    def _get_all_path(cls, parentid=None):
        # Not all sub resources require a parent id.  Eg: /api/v2/products/skus?sku=<value>
        if (parentid):
            return "%s/%s/%s" % (cls.parent_resource, parentid, cls.resource_name)
        else:
            return "%s/%s" % (cls.parent_resource, cls.resource_name)

    @classmethod
    def all(cls, parentid=None, connection=None, **params):
        response = cls._make_request('GET', cls._get_all_path(parentid), connection, params=params)
        return cls._create_object(response, connection=connection)

    @classmethod
    def iterall(cls, parentid=None, connection=None, **kwargs):
        """
            Returns a autopaging generator that yields each object returned one by one.
        """
        start_token = None
        params = kwargs.copy()
        params['max-results'] = 250

        while True:
            if start_token:
                params['start-token'] = start_token
            else:
                params.pop('start-token', None)

            rsp = cls._make_request('GET', cls._get_all_path(parentid), connection, params=params)
            if rsp:
                resources = rsp.get('resources', [])
                if not resources:
                    break

                for resource in resources:
                    yield cls._create_object(resource, connection=connection)

                # if 'kind' in rsp and rsp['kind'] == 'content#productsListResponse' and \
                #     'nextPageToken' in rsp and rsp['nextPageToken']:
                if 'nextPageToken' in rsp and rsp['nextPageToken']:
                    start_token = rsp['nextPageToken']
                    continue

                break
            else:
                yield cls._create_object({}, connection=connection)  # needed for case where there is no objects

            break


class UpdateableApiResource(ApiResource):
    def _update_path(self):
        return "%s/%s" % (self.resource_name, self.id)

    def update(self, updates):
        response = self._make_request(
            'POST', self._update_path(), self._connection, data=json.dumps(updates))
        return self._create_object(response, connection=self._connection)


class UpdateableApiSubResource(CreateableApiSubResource):
    def _update_path(self):
        return "%s/%s/%s/%s" % (self.parent_resource, self.parent_id(), self.resource_name, self.id)

    def update(self, updates):
        response = self._make_request(
            'POST', self._update_path(), self._connection, data=json.dumps(updates))
        return self._create_object(response, connection=self._connection)


class DeleteableApiResource(ApiResource):
    def _delete_path(self):
        return "%s/%s" % (self.resource_name, self.id)

    def delete(self):
        return self._make_request('DELETE', self._delete_path(), self._connection)


class DeleteableApiSubResource(ApiSubResource):
    def _delete_path(self):
        return "%s/%s/%s/%s" % (self.parent_resource, self.parent_id(), self.resource_name, self.id)

    def delete(self):
        return self._make_request('DELETE', self._delete_path(), self._connection)


class CountableApiResource(ApiResource):
    @classmethod
    def _count_path(cls):
        return "%s/count" % (cls.resource_name)

    @classmethod
    def count(cls, connection=None, **params):
        response = cls._make_request('GET', cls._count_path(), connection, params=params)
        return response['count']


class CountableApiSubResource(ApiSubResource):
    # Account for the fairly common case where the count path doesn't include the parent id
    count_resource = None

    @classmethod
    def _count_path(cls, parentid=None):
        if parentid is not None:
            return "%s/%s/%s/count" % (cls.parent_resource, parentid, cls.resource_name)
        elif cls.count_resource is not None:
            return "%s/count" % (cls.count_resource)
        else:
            # misconfiguration
            raise NotImplementedError('Count not implemented for this resource.')

    @classmethod
    def count(cls, parentid=None, connection=None, **params):
        response = cls._make_request('GET', cls._count_path(parentid), connection, params=params)
        return response['count']
